import json
import logging
import sys

from .message_handler import process_handlers
from .stamp import (SignatureStamp, TransportStamp, BusStamp, StopPropagationStamp)
from .envelope import (Envelope)
import hmac

FORMAT = '%(asctime)s %(levelname)s:%(name)s:%(message)s'
logging.basicConfig(format=FORMAT)
logger = logging.getLogger('messenger')
logger.setLevel(logging.DEBUG)

class MiddlewareInterface:
    """
    middleware permettant d'intercepter les envelopes en envoi comme en reception
    """
    def __init__(self, priority:int = 100):
       self.priority = priority

    def handle(self,envelope:Envelope, stack) -> Envelope:
        raise NotImplementedError

class EndMiddleware(MiddlewareInterface):
    """
    marque la fin des middlewares
    """
    def __init__(self):
       super().__init__()

    def handle(self,envelope:Envelope, stack) -> Envelope:
        return envelope

class SignatureMiddleware(MiddlewareInterface):
    """
    ajoute une signature electronique aux données envoyées
    """
    def __init__(self):
       super().__init__(1)

    def handle(self,envelope:Envelope, stack) -> Envelope:

        if envelope.last("SendingStamp"):
            stamp:SignatureStamp = envelope.last("SignatureStamp")
            if not stamp:
                body = str(envelope.message)
                token = hmac.digest(b"zakes25649", body.encode(), digest="sha256").hex()

                envelope = envelope.update(SignatureStamp("myProducerId",token))

        return stack.next().handle(envelope, stack)

class SendMiddleware(MiddlewareInterface):
    """
    middleware en charge de l'envoi des messages
    """
    def __init__(self):
        super().__init__(-1)

    def handle(self,envelope:Envelope, stack) -> Envelope:

        if envelope.last("SendingStamp") and not envelope.last("DispatchAfterCurrentBusStamp"):
            stamp_transport: TransportStamp = envelope.last("TransportStamp")
            transport = stamp_transport.transport
            envelope = transport.produce(envelope)


        return stack.next().handle(envelope,stack)

class MessageHandlerMiddleware(MiddlewareInterface):
    """
    middleware en charge de l'execution du code apres reception d'un message
    """
    def __init__(self):
        super().__init__(0)

    def handle(self,envelope:Envelope, stack) -> Envelope:

        if envelope.last("ReceivedStamp") and not envelope.last("SkipReceivedStamp"):
            envelope = process_handlers(envelope)

        return stack.next().handle(envelope,stack)


class DispatchAfterCurrentBusMiddleware(MiddlewareInterface):
    """
    middleware en charge de l'execution d'un bus apres la fin de traitement du bus en cours
    il faut arreter immediatement le traitement des autres middlewares
    enregistrer l'envelope et le redispatcher a la fin du bus
    """
    def __init__(self):
        super().__init__(2)

    def handle(self,envelope:Envelope, stack) -> Envelope:

        if envelope.last("DispatchAfterCurrentBusStamp") and envelope.last("SendingStamp"):
            stamp:BusStamp = envelope.last("BusStamp")
            stamp.bus.add_event(envelope)
            return envelope

        return stack.next().handle(envelope, stack)



class MiddlewareManager:
    """
    le middleware manager, il sert a orchestrer tout les middlewares
    """
    def __init__(self, middlewares: list=[]):
        self._middlewares = []
        self.currentEnvelope = None
        self.iterable = None

        ms:list = [
            DispatchAfterCurrentBusMiddleware(),
            SendMiddleware(),
            MessageHandlerMiddleware(),
        ] + middlewares

        for middleware in ms:
            if middleware.priority <= 10 and type(middleware) not in [SendMiddleware, MessageHandlerMiddleware, DispatchAfterCurrentBusMiddleware]:
                middleware.priority = 11
            self.add(middleware)

        self._middlewares = sorted(self._middlewares, key=lambda el: el.priority, reverse=True)

    def add(self, middleware: MiddlewareInterface):
        """ ajoute un middleware au manager"""
        try:
            self._middlewares.index(middleware)
        except ValueError as e:
            self._middlewares.append(middleware)
        return self

    def remove(self, middleware: MiddlewareInterface):
        """ supprime un middleware au manager"""
        try:
            self._middlewares.remove(middleware)
        except ValueError as e:
            pass
        return self

    def next(self) -> MiddlewareInterface:
        """
        retourne le prochain
        :return:
        """
        middleware = None
        try:
            # if not self.iterable:
            #     raise StopIteration
            middleware = next(self.iterable)
        except StopIteration as e:
            middleware = EndMiddleware()

        return middleware

    def run(self, envelope: Envelope) -> Envelope:
        self.currentEnvelope = envelope
        self.iterable = iter(self._middlewares)
        envelope = self.next().handle(envelope, self)

        try:
            next(self.iterable)
            envelope = envelope.update(StopPropagationStamp())
        except:
            pass
        # finally:
        #     self.iterable = None

        return envelope




class AMQPBasicProperties:

    def __init__(self):
        self.content_type = None
        self.content_encoding = None
        self.headers = {}
        self.delivery_mode = 2
        self.priority = 0
        self.correlation_id = None
        self.reply_to = None
        self.expiration = None
        self.message_id = None
        self.timestamp = 0
        self.type = None
        self.user_id = None
        self.app_id = None
        self.cluster_id = None

class StampInterface:
    pass

class NonSendableStampInterface(StampInterface):
    pass

class AmqpStamp(NonSendableStampInterface):
    def __init__(self,  routing_key:str = "",  flags:int = 2, attributes:AMQPBasicProperties = AMQPBasicProperties()):
        super().__init__()
        self.routing_key = routing_key
        self.flags = flags
        self.attributes = attributes
        self.is_retryAttempt = False

class SendingStamp(NonSendableStampInterface):
    """
    Stamp indiquant que l'envelope est en cours d'envoi
    """
    def __init__(self):
        super().__init__()

class SentStamp(NonSendableStampInterface):
    """
    Stamp indiquant que l'envelope a bien été envoyé
    """
    def __init__(self):
        super().__init__()

class NotSentStamp(NonSendableStampInterface):
    """
    Stamp indiquant que l'envelope n'a pas pu être envoyé
    """
    def __init__(self):
        super().__init__()

class ReceivedStamp(NonSendableStampInterface):
    """
    Stamp indiquant que l'envelope a bien été recu
    """
    def __init__(self):
        super().__init__()

class SkipReceivedStamp(NonSendableStampInterface):
    """
    Stamp indiquant que l'envelope doit être bypassé le middleware MessageHandlerMiddleware
    """
    def __init__(self):
        super().__init__()



class DispatchAfterCurrentBusStamp(NonSendableStampInterface):
    """
    Stamp indiquant que l'envelope doit etre correctement  traité par un handler
    avant de le passer au prochain handler
    """
    def __init__(self):
        super().__init__()

class StopPropagationStamp(NonSendableStampInterface):
    """
    Stamp indiquant que l'envelope une fin prématuré dans un middleware
    """
    def __init__(self):
        super().__init__()

class BusStamp(NonSendableStampInterface):
    """
    Stamp indiquant le bus par lequel transit le message
    """
    def __init__(self,bus):
        super(BusStamp, self).__init__()
        self.bus = bus


class TransportStamp(NonSendableStampInterface):
    """
    Stamp indiquant le transport par lequel le message est envoyé
    """
    def __init__(self,transport, attributes:dict={}):
        super(TransportStamp, self).__init__()
        self.transport = transport
        self.attributes = attributes

class ResultStamp(NonSendableStampInterface):
    """
    Stamp qui stock le resultat d'une action envoyé dans le bus
    """
    def __init__(self,result):
        super(ResultStamp, self).__init__()
        self.result = result

class SignatureStamp(StampInterface):
    """
    Stamp ajoutant la signature numeric des données transportées
    """
    def __init__(self, producerId:str, payloadToken:str):
        super(SignatureStamp, self).__init__()
        self.producerId = producerId
        self.payloadToken = payloadToken


import asyncio
import json
import logging

import pika
from .service_container import message_bus

FORMAT = '%(asctime)s %(levelname)s:%(name)s:%(message)s'
logging.basicConfig(format=FORMAT)
logger = logging.getLogger('messenger')
logger.setLevel(logging.DEBUG)

class WorkerInterface:

    def __init__(self):
        pass

    def _connect(self):
        """ etablit la connection avec le serveur AMQP"""
        raise NotImplementedError

    def consume(self):
        """ ecoute les nouveaux messages et les dispatch dans le bus"""
        raise NotImplementedError

    def _on_message(self, ch, method, properties, body):
        """ traite la reception de nouveaux message """
        raise NotImplementedError

class DefaultWorker(WorkerInterface):

    def __init__(self):
        super(DefaultWorker, self).__init__()
        self.connection = None
        self.channel = None
        self.queue_name = None
        self.exchange_name = None

    def _connect(self):
        self.connection, self.channel, self.queue_name, self.exchange_name = createConnection()
        return self

    async def consume(self):
        # result = channel.queue_declare('', exclusive=True, durable=True)
        # queue_name = result.method.queue
        max_attempts = 100

        while True:
            try:
                self._connect()
                self.channel.basic_consume(queue=self.queue_name, on_message_callback=self._on_message, auto_ack=False)

                try:
                    self.channel.start_consuming()
                except KeyboardInterrupt:
                    self.channel.stop_consuming()
                    if self.connection:
                        self.connection.close()
                    break

            except pika.exceptions.ConnectionClosedByBroker:
                # Uncomment this to make the example not attempt recovery
                # from server-initiated connection closure, including
                # when the node is stopped cleanly
                # except pika.exceptions.ConnectionClosedByBroker:
                #     pass
                logger.debug("Impossible de se connecter au serveur")

            except pika.exceptions.AMQPConnectionError as e:
                logger.debug("Impossible de se connecter au serveur")
                # Do not recover on channel errors

            except pika.exceptions.AMQPChannelError as err:
                logger.debug("Caught a channel error: {}, stopping...".format(err))
                print("Caught a channel error: {}, stopping...".format(err))

            except Exception as e:
                logger.debug(e)
                continue
            finally:
                logger.debug("Reconnection du service...")

            await asyncio.sleep(1)

            #max_attempts -= 1

    def _on_message(self,ch, method, properties, body):
        try:
            print(" [x] %r:%r" % (method.routing_key, body))
            # task = asyncio.create_task(message_bus.receive(body.decode(),properties.__dict__))
            try:
                message_bus.receive(body.decode(),properties.__dict__)
            except Exception as e:
                logger.debug(e)
            
            ch.basic_ack(delivery_tag=method.delivery_tag)
        except Exception as e:
            logger.debug(e)
            ch.basic_ack(delivery_tag=method.delivery_tag)

            try:
                message = json.loads(body.decode())
                headers = {"x-retry":True, "x-retry-count":0}

                if "x-retry-count" in properties.headers:
                    headers["x-retry-count"] = properties.headers["x-retry-count"] + 1

                message_bus.dispatch(message, method.routing_key,{"headers":headers})
            except:
                pass


import json
import logging
import pathlib
import re
from collections import namedtuple
from typing import NamedTuple

from .envelope import Envelope
from .stamp import AmqpStamp, ResultStamp

FORMAT = '%(asctime)s %(levelname)s:%(name)s:%(message)s'
logging.basicConfig(format=FORMAT)
logger = logging.getLogger('messenger')
logger.setLevel(logging.DEBUG)

_handlers = []


class CommandInterfaceMeta(type):

    def __new__(metacls, cls, bases, attrs):

        _new_dict = {"__dict__":{}}
        for k,v in attrs.items():
            if not k.startswith("__") and not k.endswith("__") and not callable(v):
                _new_dict["__dict__"][k] = v

        for k,v in _new_dict["__dict__"].items():
            del attrs[k]

        attrs.update(_new_dict)

        return type.__new__(metacls, cls, bases, attrs)


class CommandInterface(metaclass=CommandInterfaceMeta):

    def __init__(self, payload:dict={}):
        self._hydrate(payload)

    def _hydrate(self, payload:dict={}):
        for k,v in payload.items():
            if k not in self:
                raise AttributeError("'{}'".format(k))

            if getattr(self,k) is None:
                self.__dict__[k] = v

    def __eq__(self, other):
        for k,v in self.__dict__.items():
            if k not in other or v != other[k]:
                return False
        return True

    def __call__(self, payload:dict={}):
        return CommandInterface(payload)

    def __setattr__(self, key, value):
        raise Exception("{} is immutable object".format(self.__class__.__name__))

    def __getattr__(self, k):
        if k not in self.__dict__:
            raise AttributeError("'{}'".format(k))
        return self.__dict__.get(k)

    def __setitem__(self, key, value):
        self.__setattr__(key,value)

    def __getitem__(self, key):
        return self.__getattr__(key)

    def __str__(self):
        return json.dumps(self.__repr__())

    def __repr__(self):
        return self.__dict__

    def __contains__(self, k):
        return True if k in self.__dict__ else False

class DefaultCommand(CommandInterface):
    action = None
    payload = None
    def __init__(self, payload:dict={}):
        super().__init__(payload)


def handler(transport:str=None, bus:str=None, binding_key:str=None, priority:int=0, command:CommandInterface=None):
    def wrapper(fn):
        def func(*args,**kwargs):
            return fn(*args, **kwargs)

        item = {
            "transport":transport,
            "bus":bus,
            "binding_key":binding_key,
            "priority": priority,
            "command": command,
            "callback":func,
            "controller":fn
        }

        if type(fn) == type: # annotation mise sur une class
            item["type"] = "class"
        elif type(fn).__name__ == "function": # annotation mise sur une function
            item["type"] = "function"

        _handlers.append(item)
        return func
    return wrapper



def process_handlers(envelope:Envelope) -> Envelope:
    from .transport import AMQPTransport

    stamp = envelope.last("TransportStamp")
    transport = stamp.transport
    transport_attributes:dict = stamp.attributes

    stamp = envelope.last("BusStamp")
    bus = stamp.bus
    message = envelope.message

    print(transport_attributes)

    _handlers_selected = []
    for p in _handlers:

        is_ok = True
        criterias = {}
        # un check transport est actif
        if p.get("transport"):
            criterias["transport"] = False
            if p.get("transport") == transport.definition.name:
                criterias["transport"] = True

        # un check bus est actif
        if p.get("bus"):
            criterias["bus"] = False
            if p.get("bus") == bus.definition.name:
                criterias["bus"] = True

        # un check binding_key est actif
        if p.get("binding_key"):
            criterias["binding_key"] = False

            if not isinstance(message,CommandInterface):
                binding_key = message.get("action")
                if re.search(r"^{}$".format(p.get("binding_key")), binding_key):
                    criterias["binding_key"] = True


        # un check command est actif
        if p.get("command"):
            criterias["command"] = False
            if isinstance(message, CommandInterface):
                if isinstance(message,p.get("command")):
                    criterias["command"] = True
                    print(dir(p["controller"]),p["controller"].__annotations__)

        if p["type"] == "class":
            instance_ = p["controller"]()
            for name,inst in instance_.__call__.__annotations__.items():
                if isinstance(message,inst):
                    criterias["command_type_hint"] = True
                    break

        elif p["type"] == "function":
            for name,inst in p["controller"].__annotations__.items():
                if isinstance(message,inst):
                    criterias["command_type_hint"] = True
                    break

        # critères d'invalidité
        if transport_attributes.get("transport") and p.get("transport") != transport_attributes.get("transport"):
            criterias = {k:False for k,v in criterias.items()}

        if transport_attributes.get("bus") and p.get("bus") != transport_attributes.get("bus"):
            criterias = {k:False for k,v in criterias.items()}


        # check de validation globale
        if len(criterias):
            for k,v in criterias.items():
                if v == False:
                    is_ok = False
                    break

            if is_ok:
                _handlers_selected.append(p)

    _handlers_selected.sort(key=lambda i: i["priority"], reverse=True)

    for p in _handlers_selected:
        callback = p["callback"]
        properties = transport_attributes
        ret = None

        if isinstance(transport, AMQPTransport):
            amqpStamp: AmqpStamp = envelope.last("AmqpStamp")
            if amqpStamp:
                properties = amqpStamp.attributes.__dict__

        argv = [envelope.message]
        argc = 1
        if p["type"] == "class":
            callback = callback()
            argc = len(callback.__call__.__annotations__)

        elif p["type"] == "function":
            argc = len(p["controller"].__annotations__)

        if (argc >= 2):
            argv.append(properties)
        ret = callback(*argv)

        if ret:
            envelope = envelope.update(ResultStamp(ret))

    return envelope

class MessageHandlerInterface:

    def __call__(self, message):
        raise NotImplementedError

class HandlerInterface:

    def __init__(self, binding_key:str, priority:int = 0):
        self._binding_key = binding_key
        self.priority = priority

    def check(self,binding_key:str = "") -> bool:
        match = re.search(r"^{}$".format(self._binding_key), binding_key)
        return True if match else False

    def start(self,binding_key:str, payload:dict,properties:dict={}):
        if self.check(binding_key):
            return self.run(binding_key,payload,properties)
        return None

    def run(self,binding_key:str, payload:dict, properties:dict={}):
        raise NotImplementedError

    def getUsableBindingKey(self):
        binding_key = self._binding_key
        b = binding_key.split(".")
        if len(b) > 2:
            b.pop()
            binding_key = ".".join(b)
        return binding_key

class HandlerManager:

    def __init__(self, handlers:list=[]):
        self._handlers = []

        for handler in handlers:
            self._add(handler)

        self._handlers.sort(key=lambda el: el.priority, reverse=True)

    def _add(self, handler: HandlerInterface):
        try:
            self._handlers.index(handler)
        except ValueError as e:
            self._handlers.append(handler)
        return self


    def run(self,match:str, payload:dict, properties:dict={}):

        self._handlers.sort(key=lambda el: el.priority, reverse=True)

        for handler in self._handlers:
            rst = handler.start(match, payload, properties)
            if rst:
                return rst

        return None